package org.jvnet.jaxbcommons.equals.addon;

import java.io.IOException;
import java.util.Arrays;
import java.util.List;

import javax.xml.namespace.QName;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jvnet.jaxbcommons.addon.Customizations;
import org.jvnet.jaxbcommons.addon.CustomizedIgnoring;
import org.jvnet.jaxbcommons.addon.Ignoring;
import org.jvnet.jaxbcommons.addon.generator.MethodStrategy;
import org.jvnet.jaxbcommons.equals.addon.generator.EqualsStrategy;
import org.xml.sax.ErrorHandler;

import com.sun.tools.xjc.BadCommandLineException;
import com.sun.tools.xjc.CodeAugmenterEx;
import com.sun.tools.xjc.Options;
import com.sun.tools.xjc.generator.ClassContext;
import com.sun.tools.xjc.generator.GeneratorContext;
import com.sun.tools.xjc.grammar.AnnotatedGrammar;
import com.sun.tools.xjc.grammar.ClassItem;

/**
 * Add-on for {@link Object#equals(Object)} and {@link Object#hashCode()}
 * methods generation.
 * 
 * @author Aleksei Valikov
 */
public class AddOn implements CodeAugmenterEx {

	/**
	 * Logger.
	 */
	protected Log logger = LogFactory.getLog(getClass());

	private Ignoring ignoring = new CustomizedIgnoring(
			new QName[] {
					org.jvnet.jaxbcommons.equals.addon.Customizations.IGNORED_ELEMENT_NAME,
					Customizations.IGNORED_ELEMENT_NAME,
					Customizations.GENERATED_ELEMENT_NAME });

	public Ignoring getIgnoring() {
		return ignoring;
	}

	public void setIgnoring(Ignoring ignoring) {
		this.ignoring = ignoring;
	}

	/**
	 * Does nothing.
	 * 
	 * @param opt
	 *            ignored.
	 * @param args
	 *            ignored.
	 * @param i
	 *            ignored.
	 * @return Alway returns <code>0</code>.
	 * @throws com.sun.tools.xjc.BadCommandLineException
	 *             Never thrown.
	 * @throws java.io.IOException
	 *             Never thrown.
	 */
	public int parseArgument(final Options opt, final String[] args, final int i)
			throws BadCommandLineException, IOException {
		return 0; // no option recognized
	}

	public String getOptionName() {
		return "Xequals";
	}

	public String getUsage() {
		return "  -Xequals                     :  adds equals(...) method";
	}

	/**
	 * Augments the code.
	 * 
	 * @param grammar
	 *            annotated grammar.
	 * @param generatorContext
	 *            generator context.
	 * @param options
	 *            options.
	 * @param errorHandler
	 *            error handler.
	 * @return If the add-on executes successfully, return true. If it detects
	 *         some errors but those are reported and recovered gracefully,
	 *         return false.
	 */
	public boolean run(final AnnotatedGrammar grammar,
			final GeneratorContext generatorContext, final Options options,
			final ErrorHandler errorHandler) {

		logger.debug("Starting code augmentation.");

		// Process class items
		final ClassItem[] classItems = grammar.getClasses();
		logger.debug("Processing [" + classItems.length + "] class items.");
		for (int index = 0; index < classItems.length; index++) {
			final ClassItem classItem = classItems[index];
			final ClassContext classContext = generatorContext
					.getClassContext(classItem);
			logger.debug("Processing [" + classItem.name + "].");
			getEqualsStrategy().generate(classContext, null,
					classContext.implClass);
		}
		logger.debug("Finished processing class items.");
		logger.debug("Finished code augmentation.");
		return true;
	}

	public List getCustomizationURIs() {
		return Arrays
				.asList(new String[] {
						org.jvnet.jaxbcommons.equals.addon.Customizations.NAMESPACE_URI,
						Customizations.NAMESPACE_URI });
	}

	protected MethodStrategy equalsStrategy = new EqualsStrategy(getIgnoring());

	public MethodStrategy getEqualsStrategy() {
		return equalsStrategy;
	}

	public void setStrategy(MethodStrategy equalsStrategy) {
		this.equalsStrategy = equalsStrategy;
	}
}
